# Importing the Packages:
import optuna
import argparse
import joblib
import tempfile
import pandas as pd
from sklearn import linear_model
from sklearn import datasets
from sklearn import model_selection
import sklearn.ensemble as ensemble
import mlflow
import random
import os
# import sktime
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
# import xgboost as xgb
# from xgboost import XGBClassifier
from sklearn.metrics import accuracy_score, recall_score, precision_score
from sklearn.metrics import recall_score
from sklearn.pipeline import Pipeline


from sklearn.model_selection import train_test_split

from sklearn.utils import all_estimators
from sklearn.base import ClassifierMixin
# from sklearn.utils.testing import all_estimators
from sklearn.utils._testing import ignore_warnings




CLASSIFIERS = [est for est in all_estimators() if issubclass(est[1], ClassifierMixin)]


SENSORS_FEATURES = ['sensor_1','sensor_2','sensor_3','sensor_4','sensor_5',
 'sensor_6', 'sensor_7', 'sensor_8', 'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
 'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18', 'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23', 'sensor_24']

STAGES = ['baseline', 'absorb', 'pause', 'desorb', 'flush']
TARGET_COL = 'result'
   

def get_plots(model, X_test, y_pred):

    model_attributes = dir(model)


    cfm_plot = plot_confusion_matrix(model, X_test, y_pred)
    cfm_plot.figure_.savefig("Confusion Matrix.png")
    cm = os.path.join(os.getcwd(), 'Confusion Matrix.png')
    mlflow.log_artifact(cm)

    


    if 'predict_proba' in model_attributes:
        try:


            roc_plot = plot_roc_curve(model, X_test, y_pred)
            roc_plot.figure_.savefig("AUC.png")
            auc = os.path.join(os.getcwd(), 'AUC.png')
            mlflow.log_artifact(auc)

            pr_plot = plot_precision_recall_curve(model, X_test, y_pred)
            roc_plot.figure_.savefig("Precision Recall.png")
            pr = os.path.join(os.getcwd(), 'Precision Recall.png')
            mlflow.log_artifact(pr)




            # reliability diagram
            fop, mpv = calibration_curve(y_test, probs, n_bins=10, normalize=True)
            # plot perfectly calibrated
            plt.plot([0, 1], [0, 1], linestyle='--')
            # plot model reliability
            plt.plot(mpv, fop, marker='.')
            plt.savefig('Calibration Curve.png')
            cc = os.path.join(os.getcwd(), 'Calibration Curve.png')
            mlflow.log_artifact(cc)

        except:
            pass
  

    
    if 'feature_importances_' in model_attributes:  
        feat_importances = pd.Series(model.feature_importances_) #, index=X.columns
        feat_importances.nlargest(20).plot(kind='barh')
        plt.savefig('Feature Importance.png')
        fi = os.path.join(os.getcwd(), 'Feature Importance.png')
        mlflow.log_artifact(fi)




def specificity(actual, pred, pos_label=0):
    return recall_score(actual, pred, pos_label=pos_label)




def objective(trial, df, study_name, stage, sensor, use_average, model_mode, suggest, custom_features, remote_tracking):

        mlflow.set_experiment(study_name)

        # artifact_path = mlflow.get_artifact_uri()

        mlflow.start_run()

        if remote_tracking:
                mlflow.set_tracking_uri(
            # "http://ec2-13-40-214-238.eu-west-2.compute.amazonaws.com:5000/"
            # "http://ec2-13-40-86-186.eu-west-2.compute.amazonaws.com:5000/"
            "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"
            )
        else:
            mlflow.set_tracking_uri("http://localhost:5000")
    #     uri = os.path.join(os.getcwd(), "sqlite:///mlruns.db")
    #     mlflow.set_tracking_uri(uri)



    # with mlflow.start_run():
    

        X = df.drop('result', axis=1)
        y = df['result']

        model = RandomForestClassifier()
        X_train, X_test, y_train, y_test = train_test_split(X, y)
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)

        accuracy = model.score(X_test, y_test)
        f1 = f1_score(y_test, y_pred, average="macro")
        precision = precision_score(y_test, y_pred, average="macro")
        sensitivity = recall_score(y_test, y_pred, average="macro")
        specificity = recall_score(y_test, y_pred, average="macro", pos_label=0)

        metrics = { 'accuracy': accuracy, 
                    'precision': precision, 
                    'sensitivity': sensitivity,
                    'specificity': specificity,
                    'f1': f1}


       
        get_plots(model, X_test, y_pred)
        mlflow.sklearn.log_model(model, artifact_path=model_name)

            


            
        model_params = model.get_params()
        


        mlflow.log_metrics(metrics)

        try:
            mlflow.set_tag('sensor_cols', sensors_to_use)
            mlflow.set_tag('stage', stages_to_use)
            mlflow.set_tag('use_average', use_average)
        except:
            pass

        mlflow.set_tag('model_name', model_name)
        mlflow.set_tag('model_params', model_params)
        
        mlflow.set_tag('model_mode', model_mode)
        mlflow.set_tag('suggest', suggest)

        print('Click on this link to track experiments: ', mlflow.get_tracking_uri())
        mlflow.end_run()

        return sensitivity







def train_experiments(df, study_name='raptor', direction='maximize', stage='pause', sensor='12', use_average=False, model_mode='random', n_trials=5, custom_features=False, suggest=False, remote_tracking=True):
    """trains several models during different trials and logs them. Experiemnt can be tracked on "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"


    Args:
        df (pandas dataframe): dataframe of cyclic sensor data to be used for training
        study_name (str, optional): optuna study name to use. Defaults to 'raptor'.
        direction (str, optional): dirtection of objective. Defaults to 'maximize'.
        stage (str or list, optional): Can take any of ('baseline', 'absorb', 'pause', 'desorb', 'flush') as a str or combination as a list. 
                                        Can also take str 'all' for all stages and 'random' for random stage or stages. Defaults to 'pause'.        sensor (str, optional): _description_. Defaults to '12'.
        use_average (bool, optional): if True, uses average of stages selected. Defaults to False.
        model_mode (str, optional): _description_. Defaults to 'random'.
        n_trials (int, optional): number of times to run experiments. Defaults to 5.
        suggest (bool, optional): if True, suggest features, stage, model mode to use for training. Defaults to False.
    """    
    
    


    study = optuna.create_study(study_name=study_name, direction=direction)
    study.optimize(lambda trial: objective(trial, df, study_name, stage, sensor, use_average, model_mode, suggest, custom_features, remote_tracking), n_trials=n_trials)




from scipy.signal import find_peaks
from scipy import stats
import pandas as pd
import numpy as np
import xgboost as xgb
from boruta import BorutaPy
from tsfresh.utilities.dataframe_functions import impute
from tsfresh.feature_extraction import ComprehensiveFCParameters







# FEATURES = ['sensor_1', 'sensor_2',
#        'sensor_3', 'sensor_4', 'sensor_5', 'sensor_6', 'sensor_7', 'sensor_8',
#        'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
#        'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18',
#        'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23',
#        'sensor_24']

FEATURES = [ 'exp_unique_id', 'timesteps', 'sensor_1', 'sensor_2',
       'sensor_3', 'sensor_4', 'sensor_5', 'sensor_6', 'sensor_7', 'sensor_8',
       'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
       'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18',
       'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23',
       'sensor_24']

extraction_settings = ComprehensiveFCParameters()



def iqr(x):
    q3, q1 = np.percentile(x, [75 ,25])
    return q3 - q1

def avg_absolute_diff(x):
    return np.mean(np.absolute(x - np.mean(x)))

def max_min_diff(x):
    return np.max(x) - np.min(x)

def median_abs_dev(x):
     return np.median(np.absolute(x - np.median(x)))

def negative_count(x):
    return np.sum(x < 0)

def positive_count(x):
    return np.sum(x > 0)

def values_above_mean(x):
    return np.sum(x > x.mean())

def baseline(x):
    return x.iloc[:8].mean()

def avg_of_pause(x):
    return x.iloc[[37,38,39,40]].mean()

def pause_phase_area(x):
    x = x.iloc[[37,38,39,40]]
    return np.trapz(x) 


def peaks(x):
    return len(find_peaks(x)[0])


# def extract_features(df, unique_id='exp_unique_id', label='result', features=FEATURES):

#     features = features + [unique_id, label]
#     df = df[features].groupby(by=[unique_id, label]).agg([min, max, np.std, iqr, np.median, stats.skew, stats.kurtosis, \
#     np.trapz, avg_absolute_diff, peaks, values_above_mean, positive_count, negative_count, median_abs_dev, \
#     max_min_diff, baseline, avg_of_pause, pause_phase_area]).reset_index()[features]

#     df.columns = ["_".join(col) for col in df.columns.to_flat_index()]
#     df = df.rename(columns={'result_':'result'}).drop('exp_unique_id_', axis=1)
#     return df


def get_features(df, unique_id='exp_unique_id', label='result', timesteps='timesteps', features=FEATURES):

    df = df[FEATURES]
    y = df.groupby(unique_id).first()[label]

    X = extract_features(df, column_id=unique_id, column_sort=timesteps,
                     default_fc_parameters=extraction_settings,
                     # we impute = remove all NaN features automatically
                     impute_function=impute)



    return X, y

forest = xgb.XGBClassifier()
boruta = BorutaPy(
estimator = forest, 
n_estimators = 'auto',
max_iter = 100 # number of trials to perform
)

def get_relevant_features(X, y):
    ### fit Boruta (it accepts np.array, not pd.DataFrame)
    boruta.fit(np.array(X), np.array(y))
    ### print results
    green_area = X.columns[boruta.support_].to_list()
    blue_area = X.columns[boruta.support_weak_].to_list()
    X_f = X[green_area]
    dff = X_f.join(y)


    return dff, green_area
